r"""
Each tool has its own method of ignoring errors and warnings.
For example, pylint requires a comment of the form

    # pylint disable=<error codes>

PEP8 will not warn on lines with

    # noqa

Additionally, flake8 follows that convention for pyflakes errors,
but pyflakes itself does not.

Finally, an entire file is ignored by flake8 if this line is found
in the file:

    # flake8\: noqa       (the \ is needed to stop prospector ignoring this file :))

This module's job is to attempt to collect all of these methods into
a single coherent list of error suppression locations.
"""

import re
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Optional

from prospector import encoding
from prospector.blender import BLEND_COMBOS
from prospector.exceptions import FatalProspectorException
from prospector.message import Message
from prospector.tools.base import PEP8_IGNORE_LINE_CODE, ToolBase

_FLAKE8_IGNORE_FILE = re.compile(r"flake8[:=]\s*noqa", re.IGNORECASE)
_PEP8_IGNORE_LINE = re.compile(r"#\s*noqa(\s*#.*)?$", re.IGNORECASE)
_PYLINT_SUPPRESSED_MESSAGE = re.compile(r"^Suppressed \'([a-z0-9-]+)\' \(from line \d+\)$")


class Ignore:
    source: Optional[str]
    code: str

    def __init__(
        self,
        source: Optional[str],
        code: str,
    ) -> None:
        self.source = source
        self.code = code

    def __str__(self) -> str:
        return self.code if self.source is None else f"{self.source}.{self.code}"

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self}>"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Ignore):
            return False
        return self.code == value.code and self.source == value.source

    def __hash__(self) -> int:
        return hash((self.source, self.code))


def get_noqa_suppressions(file_contents: list[str]) -> tuple[bool, set[int], dict[int, set[Ignore]]]:
    """
    Finds all pep8/flake8 suppression messages

    :param file_contents:
        A list of file lines
    :return:
        A pair - the first is whether to ignore the whole file, the
        second is a set of (0-indexed) line numbers to ignore.
    """
    ignore_whole_file = False
    ignore_lines = set()
    messages_to_ignore: dict[int, set[Ignore]] = defaultdict(set)
    for line_number, line in enumerate(file_contents):
        if _FLAKE8_IGNORE_FILE.search(line):
            ignore_whole_file = True
        if _PEP8_IGNORE_LINE.search(line):
            ignore_lines.add(line_number + 1)
        else:
            noqa_match = PEP8_IGNORE_LINE_CODE.search(line)
            if noqa_match:
                prospector_ignore = noqa_match.group(1).strip().split(",")
                prospector_ignore = [elem.strip() for elem in prospector_ignore]
                for code in prospector_ignore:
                    messages_to_ignore[line_number + 1].add(Ignore(None, code))

    return ignore_whole_file, ignore_lines, messages_to_ignore


def _parse_pylint_informational(
    messages: list[Message],
) -> tuple[set[Optional[Path]], dict[Optional[Path], dict[int, list[str]]]]:
    ignore_files: set[Optional[Path]] = set()
    ignore_messages: dict[Optional[Path], dict[int, list[str]]] = defaultdict(lambda: defaultdict(list))

    for message in messages:
        if message.source == "pylint":
            if message.code == "suppressed-message":
                # this is a message indicating that a message was raised
                # by pylint but suppressed by configuration in the file
                match = _PYLINT_SUPPRESSED_MESSAGE.match(message.message)
                if not match:
                    raise FatalProspectorException(f"Could not parsed suppressed message from {message.message}")
                suppressed_code = match.group(1)
                line_dict = ignore_messages[message.location.path]
                assert message.location.line is not None
                line_dict[message.location.line].append(suppressed_code)
            elif message.code == "file-ignored":
                ignore_files.add(message.location.path)
    return ignore_files, ignore_messages


def _process_tool_ignores(
    tools_ignore: dict[Path, dict[int, set[Ignore]]],
    blend_combos_dict: dict[Ignore, set[Ignore]],
    messages_to_ignore: dict[Optional[Path], dict[int, set[Ignore]]],
) -> None:
    for path, lines_ignore in tools_ignore.items():
        for line, ignores in lines_ignore.items():
            for ignore in ignores:
                if ignore in blend_combos_dict:
                    messages_to_ignore[path][line].update(blend_combos_dict[ignore])


def get_suppressions(
    filepaths: list[Path],
    messages: list[Message],
    tools: Optional[dict[str, ToolBase]] = None,
    blending: bool = False,
    blend_combos: Optional[list[list[tuple[str, str]]]] = None,
) -> tuple[set[Optional[Path]], dict[Path, set[int]], dict[Optional[Path], dict[int, set[Ignore]]]]:
    """
    Given every message which was emitted by the tools, and the
    list of files to inspect, create a list of files to ignore,
    and a map of filepath -> line-number -> codes to ignore
    """
    tools = tools or {}
    blend_combos = blend_combos or BLEND_COMBOS
    blend_combos_dict: dict[Ignore, set[Ignore]] = defaultdict(set)
    if blending:
        for combo in blend_combos:
            ignore_combos = {Ignore(tool, code) for tool, code in combo}
            for ignore in ignore_combos:
                blend_combos_dict[ignore] |= ignore_combos

    paths_to_ignore: set[Optional[Path]] = set()
    lines_to_ignore: dict[Path, set[int]] = defaultdict(set)
    messages_to_ignore: dict[Optional[Path], dict[int, set[Ignore]]] = defaultdict(lambda: defaultdict(set))
    tools_ignore: dict[Path, dict[int, set[Ignore]]] = defaultdict(lambda: defaultdict(set))

    # First deal with 'noqa' style messages
    for filepath in filepaths:
        try:
            file_contents = encoding.read_py_file(filepath).split("\n")
        except encoding.CouldNotHandleEncoding as err:
            # TODO: this output will break output formats such as JSON
            warnings.warn(f"{err.path}: {err.__cause__}", ImportWarning, stacklevel=2)
            continue

        ignore_file, ignore_lines, file_messages_to_ignore = get_noqa_suppressions(file_contents)
        if ignore_file:
            paths_to_ignore.add(filepath)
        lines_to_ignore[filepath] |= ignore_lines
        for line, codes_ignore in file_messages_to_ignore.items():
            messages_to_ignore[filepath][line] |= codes_ignore

        if blending:
            for line_number, line_content in enumerate(file_contents):
                for tool_name, tool in tools.items():
                    tool_ignores = tool.get_ignored_codes(line_content)
                    for tool_ignore, offset in tool_ignores:
                        tools_ignore[filepath][line_number + 1 + offset].add(Ignore(tool_name, tool_ignore))

    # Ignore the blending messages
    if blending:
        _process_tool_ignores(tools_ignore, blend_combos_dict, messages_to_ignore)

    # Now figure out which messages were suppressed by pylint
    pylint_ignore_files, pylint_ignore_messages = _parse_pylint_informational(messages)
    paths_to_ignore |= pylint_ignore_files
    for pylint_filepath, line_codes in pylint_ignore_messages.items():
        for line_number, codes in line_codes.items():
            for code in codes:
                ignore = Ignore("pylint", code)
                messages_to_ignore[pylint_filepath][line_number].add(ignore)

    return paths_to_ignore, lines_to_ignore, messages_to_ignore



from pathlib import Path


class ProfileNotFound(Exception):
    def __init__(self, name: str, profile_path: list[Path]) -> None:
        super().__init__()
        self.name = name
        self.profile_path = profile_path

    def __repr__(self) -> str:
        return "Could not find profile {}; searched in {}".format(
            self.name,
            ":".join(map(str, self.profile_path)),
        )


class CannotParseProfile(Exception):
    def __init__(self, filepath: str, parse_error: Exception) -> None:
        super().__init__()
        self.filepath = filepath
        self.parse_error = parse_error

    def get_parse_message(self) -> str:
        return (
            f"{self.parse_error.problem}\n"  # type: ignore[attr-defined]
            f"  on line {self.parse_error.problem_mark.line}: "  # type: ignore[attr-defined]
            f"char {self.parse_error.problem_mark.column}"  # type: ignore[attr-defined]
        )

    def __repr__(self) -> str:
        return f"Could not parse profile found at {self.filepath} - it is not valid YAML"



# Largely inspired by https://github.com/pre-commit/identify/blob/main/identify/identify.py#L178

import errno
import os
import shlex
import string
from pathlib import Path
from typing import IO

printable = frozenset(string.printable)


def _shebang_split(line: str) -> tuple[str, ...]:
    try:
        # shebangs aren't supposed to be quoted, though some tools such as
        # setuptools will write them with quotes so we'll best-guess parse
        # with shlex first
        return tuple(shlex.split(line))
    except ValueError:
        # failing that, we'll do a more "traditional" shebang parsing which
        # just involves splitting by whitespace
        return tuple(line.split())


def _parse_nix_shebang(
    bytes_io: IO[bytes],
    cmd: tuple[str, ...],
) -> tuple[str, ...]:
    while bytes_io.read(2) == b"#!":
        next_line_b = bytes_io.readline()
        try:
            next_line = next_line_b.decode("UTF-8")
        except UnicodeDecodeError:
            return cmd

        for c in next_line:
            if c not in printable:
                return cmd

        line_tokens = _shebang_split(next_line.strip())
        for i, token in enumerate(line_tokens[:-1]):
            if token != "-i":  # noqa: S105
                continue
            # The argument to -i flag
            cmd = (line_tokens[i + 1],)
    return cmd


def _parse_shebang(bytes_io: IO[bytes]) -> tuple[str, ...]:
    """Parse the shebang from a file opened for reading binary."""
    if bytes_io.read(2) != b"#!":
        return ()
    first_line_b = bytes_io.readline()
    try:
        first_line = first_line_b.decode("UTF-8")
    except UnicodeDecodeError:
        return ()

    # Require only printable ascii
    for c in first_line:
        if c not in printable:
            return ()

    cmd = _shebang_split(first_line.strip())
    if cmd[:2] == ("/usr/bin/env", "-S"):
        cmd = cmd[2:]
    elif cmd[:1] == ("/usr/bin/env",):
        cmd = cmd[1:]

    if cmd == ("nix-shell",):
        return _parse_nix_shebang(bytes_io, cmd)

    return cmd


def parse_shebang_from_file(path: Path) -> tuple[str, ...]:
    """Parse the shebang given a file path."""
    if not path.exists():
        return ()
    if not os.access(path, os.X_OK):
        return ()

    try:
        with path.open("rb") as f:
            return _parse_shebang(f)
    except OSError as e:
        if e.errno == errno.EINVAL:
            return ()
        else:
            raise



import codecs
import os
import re
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from pep8ext_naming import NamingChecker
from pycodestyle import PROJECT_CONFIG, USER_CONFIG, BaseReport, StyleGuide, register_check

from prospector.finder import FileFinder
from prospector.message import Location, Message
from prospector.tools.base import ToolBase

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig

__all__ = ("PycodestyleTool",)


class ProspectorReport(BaseReport):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._prospector_messages: list[Message] = []

    def error(self, line_number: Optional[int], offset: int, text: str, check: str) -> None:
        code = super().error(
            line_number,
            offset,
            text,
            check,
        )
        if code is None:
            # The error pycodestyle found is being ignored, let's move on.
            return

        # Get a clean copy of the message text without the code embedded.
        text = text[5:]

        # mixed indentation (E101) is a file global message
        if code == "E101":
            line_number = None

        # Record the message using prospector's data structures.
        location = Location(
            path=self.filename,
            module=None,
            function=None,
            line=line_number,
            character=(offset + 1),
        )
        message = Message(
            # TODO: legacy output naming
            source="pycodestyle",
            code=code,
            location=location,
            message=text,
        )

        self._prospector_messages.append(message)

    def get_messages(self) -> list[Message]:
        return self._prospector_messages


class ProspectorStyleGuide(StyleGuide):
    def __init__(self, config: "ProspectorConfig", found_files: FileFinder, *args: Any, **kwargs: Any) -> None:
        self._config = config
        self._files = found_files
        self._module_paths = found_files.python_modules

        # Override the default reporter with our custom one.
        kwargs["reporter"] = ProspectorReport

        super().__init__(*args, **kwargs)

    def excluded(self, filename: str, parent: Optional[str] = None) -> bool:
        if super().excluded(filename, parent):
            return True

        # If the file survived pycodestyle's exclusion rules, check it against
        # prospector's patterns.
        fullpath = self._config.workdir / (parent or "") / filename
        if fullpath.is_dir():
            return False

        return fullpath not in self._module_paths


class PycodestyleTool(ToolBase):
    checker: Optional[ProspectorStyleGuide] = None

    def configure(
        self, prospector_config: "ProspectorConfig", found_files: FileFinder
    ) -> Optional[tuple[Optional[str], Optional[Iterable[Message]]]]:
        # figure out if we should use a pre-existing config file
        # such as setup.cfg or tox.ini
        external_config = None

        # 'none' means we ignore any external config, so just carry on
        use_config = False

        if prospector_config.use_external_config("pycodestyle"):
            use_config = True

            paths: list[Union[str, Path]] = [os.path.join(prospector_config.workdir, name) for name in PROJECT_CONFIG]
            paths.append(USER_CONFIG)
            ext_loc = prospector_config.external_config_location("pycodestyle")
            if ext_loc is not None:
                paths = [ext_loc] + paths  # type: ignore[assignment,operator]

            for conf_path in paths:
                if os.path.exists(conf_path) and os.path.isfile(conf_path):
                    # this file exists - but does it have pep8 or pycodestyle config in it?
                    # TODO: Remove this
                    header = re.compile(r"\[(pep8|pycodestyle)\]")
                    with codecs.open(str(conf_path)) as conf_file:
                        if any(header.search(line) for line in conf_file.readlines()):
                            external_config = conf_path
                            break

        # need to convert to strings rather than Path objects for compatibility with pycodestyle
        check_paths = [str(f.absolute()) for f in found_files.python_modules]

        # check_paths = [found_files.to_absolute_path(p) for p in check_paths]

        # Instantiate our custom pycodestyle checker.
        self.checker = ProspectorStyleGuide(
            config=prospector_config, paths=check_paths, found_files=found_files, config_file=external_config
        )
        if not use_config or external_config is None:
            configured_by = None
            # This means that we don't have existing config to use.
            # Make sure pycodestyle's code ignores are fully reset to zero before
            # adding prospector-flavoured configuration.
            self.checker.options.select = ()
            self.checker.options.ignore = tuple(prospector_config.get_disabled_messages("pycodestyle"))

            if "max-line-length" in prospector_config.tool_options("pycodestyle"):
                self.checker.options.max_line_length = prospector_config.tool_options("pycodestyle")["max-line-length"]
        else:
            configured_by = f"Configuration found at {external_config}"

        # if we have a command line --max-line-length argument, that
        # overrules everything
        max_line_length = prospector_config.max_line_length
        if max_line_length is not None:
            self.checker.options.max_line_length = max_line_length

        return configured_by, []

    def run(self, _: Any) -> list[Message]:
        assert self.checker is not None
        report = self.checker.check_files()
        return report.get_messages()


# Load pep8ext_naming into pycodestyle's configuration.
register_check(NamingChecker)



from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from vulture import Vulture

from prospector.encoding import CouldNotHandleEncoding, read_py_file
from prospector.finder import FileFinder
from prospector.message import Location, Message, make_tool_error_message
from prospector.tools.base import ToolBase

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig


class ProspectorVulture(Vulture):
    def __init__(self, found_files: FileFinder) -> None:
        Vulture.__init__(self, verbose=False)
        self._files = found_files
        self._internal_messages: list[Message] = []
        self.file: Optional[Path] = None
        self.filename: Optional[Path] = None

    def scavenge(self, _: Any = None, __: Any = None) -> None:
        # The argument is a list of paths, but we don't care
        # about that as we use the found_files object. The
        # argument is here to explicitly acknowledge that we
        # are overriding the Vulture.scavenge method.
        for module in self._files.python_modules:
            try:
                module_string = read_py_file(module)
            except CouldNotHandleEncoding as err:
                self._internal_messages.append(
                    make_tool_error_message(
                        module,
                        "vulture",
                        "V000",
                        message=(
                            f"Could not handle the encoding of this file: {err.encoding}"  # type: ignore[attr-defined]
                        ),
                    )
                )
                continue
            self.file = module
            self.filename = module
            try:
                self.scan(module_string, filename=module)
            except TypeError:
                self.scan(module_string)

    def get_messages(self) -> list[Message]:
        all_items = (
            ("unused-function", "Unused function %s", self.unused_funcs),
            ("unused-property", "Unused property %s", self.unused_props),
            ("unused-variable", "Unused variable %s", self.unused_vars),
            ("unused-attribute", "Unused attribute %s", self.unused_attrs),
        )

        vulture_messages = []
        for code, template, items in all_items:
            for item in items:
                try:
                    filename = item.file
                except AttributeError:
                    filename = item.filename
                lineno = item.lineno if hasattr(item, "lineno") else item.first_lineno
                loc = Location(filename, None, None, lineno, -1)
                message_text = template % item
                message = Message("vulture", code, loc, message_text)
                vulture_messages.append(message)

        return self._internal_messages + vulture_messages


class VultureTool(ToolBase):
    def __init__(self) -> None:
        ToolBase.__init__(self)
        self._vulture = None
        self.ignore_codes: list[str] = []

    def configure(  # pylint: disable=useless-return
        self, prospector_config: "ProspectorConfig", found_files: FileFinder
    ) -> Optional[tuple[Optional[str], Optional[Iterable[Message]]]]:
        self.ignore_codes = prospector_config.get_disabled_messages("vulture")
        return None

    def run(self, found_files: FileFinder) -> list[Message]:
        vulture = ProspectorVulture(found_files)
        vulture.scavenge()
        return [message for message in vulture.get_messages() if message.code not in self.ignore_codes]



from collections import defaultdict
from pathlib import Path
from typing import Optional

from prospector.formatters.text import TextFormatter
from prospector.message import Message

__all__ = ("GroupedFormatter",)


class GroupedFormatter(TextFormatter):
    def render_messages(self) -> str:
        output = [
            "Messages",
            "========",
            "",
        ]

        groups: dict[Path, dict[Optional[int], list[Message]]] = defaultdict(lambda: defaultdict(list))

        for message in self.messages:
            groups[self._make_path(message.location)][message.location.line].append(message)

        for filename in sorted(groups.keys()):
            output.append(str(filename))

            for line in sorted(groups[filename].keys(), key=lambda x: 0 if x is None else int(x)):
                output.append(f"  Line: {line}")

                for message in groups[filename][line]:
                    output.append(
                        "    {}: {} / {}{}".format(
                            message.source,
                            message.code,
                            message.message,
                            (f" (col {message.location.character})") if message.location.character else "",
                        )
                    )

            output.append("")

        return "\n".join(output)



from prospector.formatters.base_summary import SummaryFormatter
from prospector.message import Message

__all__ = ("TextFormatter",)


class TextFormatter(SummaryFormatter):
    def render_message(self, message: Message) -> str:
        output = []

        if message.location.module:
            output.append(f"{message.location.module} ({self._make_path(message.location)}):")
        else:
            output.append(f"{self._make_path(message.location)}:")

        output.append(
            "    L{}:{} {}: {} - {}".format(
                message.location.line or "-",
                message.location.character if message.location.character else "-",
                message.location.function,
                message.source,
                message.code,
            )
        )

        output.append(f"    {message.message}")
        ci_annotation = self.get_ci_annotation(message)
        if ci_annotation:
            output.append(ci_annotation)

        return "\n".join(output)

    def render_messages(self) -> str:
        output = [
            "Messages",
            "========",
            "",
        ]

        for message in self.messages:
            output.append(self.render_message(message))
            output.append("")

        return "\n".join(output)

    def render(self, summary: bool = True, messages: bool = True, profile: bool = False) -> str:
        output = []
        if messages and self.messages:  # if there are no messages, don't render an empty header
            output.append(self.render_messages())
        if profile:
            output.append(self.render_profile())
        if summary:
            output.append(self.render_summary())

        return "\n\n\n".join(output) + "\n"



from io import StringIO

from pylint.exceptions import UnknownMessageError
from pylint.message import Message as PylintMessage
from pylint.message import MessageDefinitionStore
from pylint.reporters import BaseReporter

from prospector.message import Location, Message


class Collector(BaseReporter):
    name = "collector"

    def __init__(self, message_store: MessageDefinitionStore) -> None:
        BaseReporter.__init__(self, output=StringIO())
        self._message_store = message_store
        self._messages: list[Message] = []

    def handle_message(self, msg: PylintMessage) -> None:
        loc = Location(msg.abspath, msg.module, msg.obj, msg.line, msg.column, msg.end_line, msg.end_column)

        # At this point pylint will give us the code but we want the
        # more user-friendly symbol
        try:
            msg_data = self._message_store.get_message_definitions(msg.msg_id)
        except UnknownMessageError:
            # this shouldn't happen, as all pylint errors should be
            # in the message store, but just in case we'll fall back
            # to using the code.
            msg_symbol = msg.msg_id
        else:
            msg_symbol = msg_data[0].symbol

        message = Message(
            "pylint",
            msg_symbol,
            loc,
            msg.msg,
            doc_url=f"https://pylint.readthedocs.io/en/latest/user_guide/messages/{msg.category}/{msg.symbol}.html",
        )
        self._messages.append(message)

    def get_messages(self) -> list[Message]:
        return self._messages



import json
import re
from multiprocessing import Process, Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
)

import mypy.api
import mypy.build
import mypy.errors
import mypy.fscache
import mypy.main

from prospector.finder import FileFinder
from prospector.message import Location, Message
from prospector.tools import ToolBase
from prospector.tools.exceptions import BadToolConfig

if TYPE_CHECKING:
    from prospector.config import ProspectorConfig

_IGNORE_RE = re.compile(r"#\s*type:\s*ignore\[([^#]*[^# ])\](\s*#.*)?$", re.IGNORECASE)

__all__ = ("MypyTool",)


def format_message(message: str) -> Message:
    character: Optional[int]
    try:
        (path, line_str, char_str, err_type, err_msg) = message.split(":", 4)
        line = int(line_str)
        character = int(char_str)
    except ValueError:
        try:
            (path, line_str, err_type, err_msg) = message.split(":", 3)
            line = int(line_str)
            character = None
        except ValueError:
            (path, err_type, err_msg) = message.split(":", 2)
            line = 0
            character = None
    location = Location(
        path=path,
        module=None,
        function=None,
        line=line,
        character=character,
    )
    return Message(
        source="mypy",
        code=err_type.lstrip(" "),
        location=location,
        message=err_msg.lstrip(" "),
    )


def _run_in_subprocess(
    q: "Queue[tuple[str, str]]", cmd: Callable[[list[str]], tuple[str, str]], paths: list[str]
) -> None:
    """
    This function exists only to be called by multiprocessing.Process as using
    lambda is forbidden
    """
    q.put(cmd(paths))


class MypyTool(ToolBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.checker = mypy.api
        self.options = ["--show-column-numbers", "--no-error-summary"]
        self.use_dmypy = False
        self.fscache = mypy.fscache.FileSystemCache()

    def configure(self, prospector_config: "ProspectorConfig", _: Any) -> None:
        options = prospector_config.tool_options("mypy")

        self.use_dmypy = options.pop("use-dmypy", False)

        # For backward compatibility
        if "follow-imports" not in options:
            options["follow-imports"] = "normal"
        if "python-2-mode" in options and "py2" not in options:
            options["py2"] = options.pop("python-2-mode")

        for name, value in options.items():
            if value is False:
                continue
            if value is True:
                self.options.append(f"--{name}")
                continue
            if isinstance(value, (int, float, str)):
                self.options.append(f"--{name}={value}")
                continue
            if isinstance(value, list):
                for v in value:
                    self.options.append(f"--{name}-{v}")
                continue

            raise BadToolConfig("mypy", f"The option {name} has an unsupported value type: {type(value)}")

        for code in prospector_config.get_disabled_messages("mypy"):
            self.options.append(f"--disable-error-code={code}")
        for code in prospector_config.get_enabled_messages("mypy"):
            self.options.append(f"--enable-error-code={code}")

    def run(self, found_files: FileFinder) -> list[Message]:
        args = [str(path) for path in found_files.python_modules]
        args.extend(self.options)
        if self.use_dmypy:
            # Due to dmypy messing with stdout/stderr we call it in a separate
            # process
            q: Queue[str] = Queue(1)
            p = Process(target=_run_in_subprocess, args=(q, self.checker.run_dmypy, ["run", "--"] + args))
            p.start()
            result = q.get()
            p.join()

            report, _ = result[0], result[1:]  # noqa
            return [format_message(message) for message in report.splitlines()]
        else:
            return self._run_std(args)

    def _run_std(self, args: list[str]) -> list[Message]:
        sources, options = mypy.main.process_options(args, fscache=self.fscache)
        options.output = "json"
        res = mypy.build.build(sources, options, fscache=self.fscache)

        messages = []
        for mypy_json in res.errors:
            mypy_message = json.loads(mypy_json)
            message = f"{mypy_message['message']}."
            if mypy_message.get("hint", ""):
                message = f"{message} {mypy_message['hint']}."
            code = mypy_message["code"]
            messages.append(
                Message(
                    "mypy",
                    code=code,
                    location=Location(
                        path=mypy_message["file"],
                        module=None,
                        function=None,
                        line=mypy_message["line"],
                        character=mypy_message["column"],
                    ),
                    message=message,
                    doc_url=f"{mypy.errors.BASE_RTD_URL}-{code}",
                )
            )

        return messages

    def get_ignored_codes(self, line: str) -> list[tuple[str, int]]:
        match = _IGNORE_RE.search(line)
        if match:
            return [(e.strip(), 0) for e in match.group(1).split(",")]
        return []



import sys
from io import TextIOWrapper
from typing import Optional


class CaptureStream(TextIOWrapper):
    def __init__(self, tty: bool) -> None:
        self.contents = ""
        self._tty = tty

    def write(self, text: str, /) -> int:
        self.contents += text
        return len(text)

    def close(self) -> None:
        pass

    def flush(self) -> None:
        pass

    def isatty(self) -> bool:
        return self._tty


class CaptureOutput:
    _prev_streams = None
    stdout: Optional[CaptureStream] = None
    stderr: Optional[CaptureStream] = None

    def __init__(self, hide: bool) -> None:
        self.hide = hide

    def __enter__(self) -> "CaptureOutput":
        if self.hide:
            is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

            self._prev_streams = (
                sys.stdout,
                sys.stderr,
                sys.__stdout__,
                sys.__stderr__,
            )
            self.stdout = CaptureStream(is_a_tty)
            self.stderr = CaptureStream(is_a_tty)
            sys.stdout, sys.__stdout__ = self.stdout, self.stdout  # type: ignore[misc]
            sys.stderr, sys.__stderr__ = self.stderr, self.stderr  # type: ignore[misc]
        return self

    def get_hidden_stdout(self) -> str:
        return "" if self.stdout is None else self.stdout.contents

    def get_hidden_stderr(self) -> str:
        return "" if self.stderr is None else self.stderr.contents

    def __exit__(self, exc_type: type, exc_val: Exception, exc_tb: type) -> None:
        if self.hide:
            assert self._prev_streams is not None
            sys.stdout, sys.stderr, sys.__stdout__, sys.__stderr__ = self._prev_streams  # type: ignore[misc]
            del self._prev_streams



import os
import re
import sys
from pathlib import Path
from typing import Any, Callable, Optional, Union

import setoptconf.config

from prospector.finder import FileFinder

try:  # Python >= 3.11
    import re._constants as sre_constants
except ImportError:
    import sre_constants  # pylint: disable=deprecated-module

import contextlib

from prospector import tools
from prospector.autodetect import autodetect_libraries
from prospector.compat import is_relative_to
from prospector.config import configuration as cfg
from prospector.message import Message
from prospector.profiles import AUTO_LOADED_PROFILES
from prospector.profiles.profile import BUILTIN_PROFILE_PATH, CannotParseProfile, ProfileNotFound, ProspectorProfile
from prospector.tools import DEFAULT_TOOLS, DEPRECATED_TOOL_NAMES


class ProspectorConfig:
    # There are several methods on this class which could technically
    # be functions (they don't use the 'self' argument) but that would
    # make this module/class a bit ugly.
    # Also the 'too many instance attributes' warning is ignored, as this
    # is a config object and its sole purpose is to hold many properties!

    def __init__(self, workdir: Optional[Path] = None):
        self.config, self.arguments = self._configure_prospector()
        self.paths = self._get_work_path(self.config, self.arguments)
        self.explicit_file_mode = all(p.is_file for p in self.paths)
        self.workdir = workdir or Path.cwd()

        self.profile, self.strictness = self._get_profile(self.workdir, self.config)
        self.libraries = self._find_used_libraries(self.config, self.profile)
        self.tools_to_run = self._determine_tool_runners(self.config, self.profile)
        self.ignores = self._determine_ignores(self.config, self.profile, self.libraries)
        self.configured_by: dict[str, Optional[Union[str, Path]]] = {}
        self.messages: list[Message] = []

    def make_exclusion_filter(self) -> Callable[[Path], bool]:
        # Only close over the attributes required by the filter, rather
        # than the entire self, because ProspectorConfig can't be pickled
        # because of the config attribute, which would break parallel
        # pylint.
        ignores, workdir = self.ignores, self.workdir

        def _filter(path: Path) -> bool:
            for ignore in ignores:
                # first figure out where the path is, relative to the workdir
                # ignore-paths/patterns will usually be relative to a repository
                # root or the CWD, but the path passed to prospector may not be
                path = path.resolve().absolute()
                if is_relative_to(path, workdir):
                    path = path.relative_to(workdir)
                if ignore.match(str(path)):
                    return True
            return False

        return _filter

    def get_tools(self, found_files: FileFinder) -> list[tools.ToolBase]:
        self.configured_by = {}
        runners = []
        for tool_name in self.tools_to_run:
            tool = tools.TOOLS[tool_name]()
            config_result = tool.configure(self, found_files)
            messages: list[Message] = []
            configured_by = None
            if config_result is not None:
                configured_by, config_messages = config_result
                if config_messages is not None:
                    messages = list(config_messages)

            self.configured_by[tool_name] = configured_by
            self.messages += messages
            runners.append(tool)
        return runners

    def replace_deprecated_tool_names(self) -> list[str]:
        # pep8 was renamed pycodestyle ; pep257 was renamed pydocstyle
        # for backwards compatibility, these have been deprecated but will remain until prospector v2
        deprecated_found = []
        replaced = []
        for tool_name in self.tools_to_run:
            if tool_name in DEPRECATED_TOOL_NAMES:
                replaced.append(DEPRECATED_TOOL_NAMES[tool_name])
                deprecated_found.append(tool_name)
            else:
                replaced.append(tool_name)
        self.tools_to_run = replaced
        return deprecated_found

    def get_output_report(self) -> list[tuple[str, list[str]]]:
        # Get the output formatter
        output_report: list[tuple[str, list[str]]]
        if self.config.output_format is not None:
            output_report = self.config.output_format
        else:
            output_report = [(self.profile.output_format, self.profile.output_target)]  # type: ignore[list-item]

        for index, report in enumerate(output_report):
            if not all(report):
                output_report[index] = (report[0] or "grouped", report[1] or [])

        return output_report

    def _configure_prospector(self) -> tuple[setoptconf.config.Configuration, dict[str, str]]:
        # first we will configure prospector as a whole
        mgr = cfg.build_manager()
        config = mgr.retrieve(*cfg.build_default_sources())
        return config, mgr.arguments

    def _get_work_path(self, config: setoptconf.config.Configuration, arguments: dict[str, str]) -> list[Path]:
        # Figure out what paths we're prospecting
        if config["path"]:
            paths = [Path(self.config["path"])]
        elif arguments["checkpath"]:
            paths = [Path(p) for p in arguments["checkpath"]]
        else:
            paths = [Path.cwd()]
        return [p.resolve() for p in paths]

    def _get_profile(
        self, workdir: Path, config: setoptconf.config.Configuration
    ) -> tuple[ProspectorProfile, Optional[str]]:
        # Use the specified profiles
        profile_provided = False
        if len(config.profiles) > 0:
            profile_provided = True
        cmdline_implicit = []

        # if there is a '.prospector.ya?ml' or a '.prospector/prospector.ya?ml' or equivalent landscape config
        # file then we'll include that
        profile_name: Union[None, str, Path] = None
        if not profile_provided:
            for possible_profile in AUTO_LOADED_PROFILES:
                prospector_yaml = os.path.join(workdir, possible_profile)
                if os.path.exists(prospector_yaml) and os.path.isfile(prospector_yaml):
                    profile_provided = True
                    profile_name = possible_profile
                    break

        strictness = None

        if profile_provided:
            if profile_name is None:
                profile_name = config.profiles[0]
                extra_profiles = config.profiles[1:]
            else:
                extra_profiles = config.profiles

            strictness = "from profile"
        else:
            # Use the preconfigured prospector profiles
            profile_name = "default"
            extra_profiles = []

        if config.doc_warnings is not None and config.doc_warnings:
            cmdline_implicit.append("doc_warnings")
        if config.test_warnings is not None and config.test_warnings:
            cmdline_implicit.append("test_warnings")
        if config.no_style_warnings is not None and config.no_style_warnings:
            cmdline_implicit.append("no_pep8")
        if config.full_pep8 is not None and config.full_pep8:
            cmdline_implicit.append("full_pep8")
        if config.member_warnings is not None and config.member_warnings:
            cmdline_implicit.append("member_warnings")

        # Use the strictness profile only if no profile has been given
        if config.strictness is not None and config.strictness:
            cmdline_implicit.append(f"strictness_{config.strictness}")
            strictness = config.strictness

        # the profile path is
        #   * anything provided as an argument
        #   * a directory called .prospector in the check path
        #   * the check path
        #   * prospector provided profiles
        profile_path = [Path(path).absolute() for path in config.profile_path]

        prospector_dir = workdir / ".prospector"
        if os.path.exists(prospector_dir) and os.path.isdir(prospector_dir):
            profile_path.append(prospector_dir)

        profile_path.append(workdir)
        profile_path.append(BUILTIN_PROFILE_PATH)

        try:
            forced_inherits = cmdline_implicit + extra_profiles
            profile = ProspectorProfile.load(profile_name, profile_path, forced_inherits=forced_inherits)
        except CannotParseProfile as cpe:
            sys.stderr.write(
                "\n".join(
                    [
                        "Failed to run:",
                        f"Could not parse profile {cpe.filepath} as it is not valid YAML",
                        f"{cpe.get_parse_message()}",
                        "",
                    ]
                )
            )
            sys.exit(1)
        except ProfileNotFound as nfe:
            search_path = ":".join(map(str, nfe.profile_path))
            module_name = nfe.name.split(":")[0]
            sys.stderr.write(
                f"""Failed to run:
Could not find profile {nfe.name}.
Search path: {search_path}, or in module 'prospector_profile_{module_name}'
"""
            )
            sys.exit(1)
        else:
            return profile, strictness

    def _find_used_libraries(self, config: setoptconf.config.Configuration, profile: ProspectorProfile) -> list[str]:
        libraries = []

        # Bring in adaptors that we automatically detect are needed
        if config.autodetect and profile.autodetect is True:
            for found_dep in autodetect_libraries(self.workdir):
                libraries.append(found_dep)

        # Bring in adaptors for the specified libraries
        for name in set(config.uses + profile.uses):
            if name not in libraries:
                libraries.append(name)

        return libraries

    def _determine_tool_runners(self, config: setoptconf.config.Configuration, profile: ProspectorProfile) -> list[str]:
        if config.tools is None:
            # we had no command line settings for an explicit list of
            # tools, so we use the defaults
            to_run = set(DEFAULT_TOOLS)
            # we can also use any that the profiles dictate
            for tool in tools.TOOLS:
                if profile.is_tool_enabled(tool):
                    to_run.add(tool)
        else:
            to_run = set(config.tools)
            # profiles have no say in the list of tools run when
            # a command line is specified

        for tool in config.with_tools:
            to_run.add(tool)

        for tool in config.without_tools:
            if tool in to_run:
                to_run.remove(tool)

        # if config.tools is None and len(config.with_tools) == 0 and len(config.without_tools) == 0:
        for tool in tools.TOOLS:
            enabled = profile.is_tool_enabled(tool)
            if enabled is None:
                enabled = tool in DEFAULT_TOOLS
            if tool in to_run and not enabled and tool not in config.with_tools and tool not in (config.tools or []):
                # if this is not enabled in a profile but is asked for in a command line arg, we keep it, otherwise
                # remove it from the list to run
                to_run.remove(tool)

        return sorted(list(to_run))

    def _determine_ignores(
        self, config: setoptconf.config.Configuration, profile: ProspectorProfile, libraries: list[str]
    ) -> list[re.Pattern[str]]:
        # Grab ignore patterns from the options
        ignores = []
        for pattern in config.ignore_patterns + profile.ignore_patterns:
            if pattern is None:
                # this can happen if someone has a profile with an empty ignore-patterns value, eg:
                #
                #  ignore-patterns:
                #  uses: django
                continue
            with contextlib.suppress(sre_constants.error):
                ignores.append(re.compile(pattern))

        # Convert ignore paths into patterns
        boundary = r"(^|/|\\)%s(/|\\|$)"
        for ignore_path in config.ignore_paths + profile.ignore_paths:
            ignore_path = str(ignore_path)
            if ignore_path.endswith("/") or ignore_path.endswith("\\"):
                ignore_path = ignore_path[:-1]
            ignores.append(re.compile(boundary % re.escape(ignore_path)))

        # some libraries have further automatic ignores
        if "django" in libraries:
            ignores += [re.compile("(^|/)(south_)?migrations(/|$)")]

        return ignores

    def get_summary_information(self) -> dict[str, Any]:
        return {
            "libraries": self.libraries,
            "strictness": self.strictness,
            "profiles": ", ".join(self.profile.list_profiles()),
            "tools": self.tools_to_run,
        }

    def exit_with_zero_on_success(self) -> bool:
        return self.config.zero_exit

    def get_disabled_messages(self, tool_name: str) -> list[str]:
        return self.profile.get_disabled_messages(tool_name)

    def get_enabled_messages(self, tool_name: str) -> list[str]:
        return self.profile.get_enabled_messages(tool_name)

    def use_external_config(self, _: Any) -> bool:
        # Currently there is only one single global setting for whether to use
        # global config, but this could be extended in the future
        return not self.config.no_external_config

    def tool_options(self, tool_name: str) -> dict[str, Any]:
        tool = getattr(self.profile, tool_name, None)
        if tool is None:
            return {}
        return tool.get("options", {})

    def external_config_location(self, tool_name: str) -> Optional[Path]:
        return getattr(self.config, f"{tool_name}_config_file", None)

    @property
    def die_on_tool_error(self) -> bool:
        return self.config.die_on_tool_error

    @property
    def summary_only(self) -> bool:
        return self.config.summary_only

    @property
    def messages_only(self) -> bool:
        return self.config.messages_only

    @property
    def quiet(self) -> bool:
        return self.config.quiet

    @property
    def blending(self) -> bool:
        return self.config.blending

    @property
    def absolute_paths(self) -> bool:
        return self.config.absolute_paths

    @property
    def max_line_length(self) -> int:
        return self.config.max_line_length

    @property
    def include_tool_stdout(self) -> bool:
        return self.config.include_tool_stdout

    @property
    def direct_tool_stdout(self) -> bool:
        return self.config.direct_tool_stdout

    @property
    def show_profile(self) -> bool:
        return self.config.show_profile

    @property
    def legacy_tool_names(self) -> bool:
        return self.config.legacy_tool_names
